/*
 * Client code generation.
 *
 * Copyright (C) 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include "client.h"
#include "common.h"
#include "declaration.h"
#include "message.h"
#include "templates.h"



static void write_signature_end(FILE *fp)
{
  fputs(";\n", fp);
}

/* Generate operation signature details for client use. */

static void _write_client_signature(struct signature *sig, FILE *fp, struct interface *iface)
{
  char *opname = get_operation_name(iface, sig);

  /* NOTE: Qualifier discarded for now. */
  /* Introduce an initial parameter for the endpoint. */

  fprintf(fp, "\nlong %s(%s _endp", opname, L4_CAP_TYPE);
  write_parameter_separator(sig->parameters, fp);
  write_parameters(sig->parameters, fp, SIGNATURE_ROLE);
  fputs(")", fp);

  free(opname);
}

/* Generate each operation signature or "function prototype" for client use. */

void write_client_signature(struct signature *sig, FILE *fp, struct interface *iface)
{
  _write_client_signature(sig, fp, iface);
  write_signature_end(fp);
}

/* Generate function source code for each operation signature. */

void write_client_function(struct signature *sig, FILE *fp, struct interface *iface)
{
  _write_client_signature(sig, fp, iface);
  fputs("\n{\n", fp);
  write_client_function_body(sig->parameters, fp, iface, sig);
  fputs("}\n", fp);
}

/* Generate a function body corresponding to an operation for client use. */

void write_client_function_body(struct parameter *param, FILE *fp,
                                struct interface *iface, struct signature *sig)
{
  int input_words, input_items, output_words, output_items;

  /* With a protocol attribute, the opcode is the attribute value. Otherwise,
     it is the operation code. */

  char *opname = get_operation_name(iface, sig),
       *protocol = get_protocol(iface->attributes),
       *opcode = get_opcode_identifier(protocol, opname);

  /* Generate the prologue. */

  fputs(client_function_body_prologue, fp);

  /* Count the number of input words, items and output words. */

  count_parameters(param, &input_words, &input_items, &output_words, &output_items);

  /* Adjust the input words where a protocol is used as the opcode. */

  if (protocol != NULL)
    input_words++;

  /* Generate types parameterised with the qualified operation name. */

  write_accessor_declaration(input_words, output_words, opname, fp);

  /* Initialise the message. */

  fprintf(fp, "\n  ipc_message_new(&msg);\n");

  /* Generate expected output item requirements. */

  if (output_items)
    fprintf(fp, "  err = ipc_message_expect(&msg, %d);\n"
                "  if (err)\n"
                "    return err;\n", output_items);

  /* Populate input parameters in the message. Dereference function parameters
     acting as "inout" parameters. */

  if (input_words)
  {
    /* Reserve space for the words before any items, obtaining a pointer to the
       words. */

    write_accessor_initialisation(IN_PARAMETER, CLIENT_ROLE, opname, fp);

    if (protocol != NULL)
      fprintf(fp, "  in_words->_op = %s;\n", opname);

    write_message_access(param, CLIENT_ROLE, IN_PARAMETER, WORD_CLASS, fp);
  }

  /* Do the same for items. */

  if (input_items)
  {
    fputs("\n", fp);
    write_message_access(param, CLIENT_ROLE, IN_PARAMETER, ITEM_CLASS, fp);
  }

  /* Send the request. */

  fprintf(fp, client_function_body_call, opcode);

  /* Retrieve output parameters from the message. */

  if (output_words)
  {
    write_accessor_initialisation(OUT_PARAMETER, CLIENT_ROLE, opname, fp);
    write_message_access(param, CLIENT_ROLE, OUT_PARAMETER, WORD_CLASS, fp);
  }

  /* Do the same for items. */

  if (output_items)
  {
    fputs("\n", fp);
    write_message_access(param, CLIENT_ROLE, OUT_PARAMETER, ITEM_CLASS, fp);
  }

  /* Return success. */

  fprintf(fp, "\n  return L4_EOK;\n");

  /* Free allocated strings. */

  free(opname);

  if (protocol == NULL)
    free(opcode);
}
