/*
 * Common routines.
 *
 * Copyright (C) 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <libgen.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"



/* Generate an error message for file opening failure. */

void open_error(const char *dir, const char *name)
{
  fprintf(stderr, "Could not open source file for writing in directory %s (name %s).\n", dir, name);
}

/* Obtain a new basename string. */

char *make_basename(const char *path)
{
  char s[strlen(path) + 1];

  strcpy(s, path);
  return strdup(basename(s));
}

/* Obtain a new directory name string. */

char *make_dirname(const char *path)
{
  char s[strlen(path) + 1];

  strcpy(s, path);
  return strdup(dirname(s));
}

/* Obtain a file pointer using a parameterised filename. */

FILE *get_output_file(const char *filename_template, const char *dir, const char *name)
{
  FILE *fp;
  char *filename;

  if (asprintf(&filename, filename_template, dir, name) == -1)
    fp = NULL;
  else
  {
    fp = fopen(filename, "w");
    free(filename);
  }

  /* Write an error message upon failure. */

  if (fp == NULL)
    open_error(dir, name);

  return fp;
}

/* Find an attribute value. */

char *get_attribute_value(struct attribute *attr, const char *name)
{
  if (attr == NULL)
    return NULL;

  if (!strcmp(attr->attribute, name) && (attr->identifiers != NULL))
    return attr->identifiers->identifier;

  return get_attribute_value(attr->tail, name);
}

/* Return any opcode attribute value. */

char *get_opcode(struct attribute *attr)
{
  return get_attribute_value(attr, "opcode");
}

/* Return any protocol attribute value. */

char *get_protocol(struct attribute *attr)
{
  return get_attribute_value(attr, "protocol");
}

/* Generate a qualified operation name for server interface operations. */

char *get_interface_operation_name(struct interface *iface,
                                   struct signature *sig, enum language lang)
{
  char *opname;

  switch (lang)
  {
    case C_LANGUAGE:
      if (asprintf(&opname, "%s_%s", iface->name, sig->operation) == -1)
        opname = NULL;
      break;

    case CPP_LANGUAGE:
      opname = strdup(sig->operation);
      break;

    default:
      opname = NULL;
      break;
  }

  return opname;
}

/* Generate a qualified operation name for structure naming and possibly opcode
   naming. */

char *get_operation_name(struct interface *iface, struct signature *sig)
{
  char *opname;

  if (asprintf(&opname, "%s_%s", iface->name, sig->operation) == -1)
    opname = NULL;

  return opname;
}

/* Return the invocation opcode given protocol and operation name details. */

char *get_opcode_identifier(char *protocol, char *opname)
{
  char *opcode;

  if (protocol != NULL)
    return protocol;

  if (asprintf(&opcode, "opcode_%s", opname) == -1)
    return NULL;

  return opcode;
}

/* Return the parameter name, being the final identifier in a list. */

char *get_parameter_name(struct identifier *ident)
{
  if (ident->tail != NULL)
    return get_parameter_name(ident->tail);
  else
    return ident->identifier;
}

/* Count the different kinds of parameters in terms of inputs, outputs, words
   and items. */

void count_parameters(struct parameter *param, int *input_words,
                      int *input_items, int *output_words, int *output_items)
{
  for (*input_words = 0, *input_items = 0, *output_words = 0, *output_items = 0;
       param != NULL;
       param = param->tail)
  {
    if (param->specifier & IN_PARAMETER)
    {
      if (param->cls & ITEM_CLASS) (*input_items)++;
      else (*input_words)++;
    }
    else
    {
      if (param->cls & ITEM_CLASS) (*output_items)++;
      else (*output_words)++;
    }
  }
}

/* Return the maximum number of input items accepted by any one of the given
   signatures. */

int get_max_input_items(struct signature *sig)
{
  int input_words, input_items, output_words, output_items;
  int max_input_items = 0;

  for (; sig != NULL; sig = sig->tail)
  {
    count_parameters(sig->parameters, &input_words, &input_items, &output_words, &output_items);

    if (input_items > max_input_items)
      max_input_items = input_items;
  }

  return max_input_items;
}

/* Return any operator needed to access the given parameter as a local name. */

char *access_name(struct parameter *param, enum component_role component)
{
  /* In the client, employ the dereferencing operator for "inout" and "out"
     word parameters since the parameters will employ pointer types. */

  if ((component == CLIENT_ROLE) && (param->specifier & OUT_PARAMETER) &&
      (param->cls & WORD_CLASS))
    return "*";

  /* In the server, when importing items from the message, the address of the
     variable to hold the capability value must be used. */

  if ((component == SERVER_ROLE) && (param->specifier & IN_PARAMETER) &&
      (param->cls & ITEM_CLASS))
    return "&";

  return "";
}

/* Return the item type name for the given parameter. */

char *item_type_name(struct parameter *param)
{
  return param->cls == FPAGE_ITEM ? "fpage" : "capability";
}

/* When copying output items from the message (in the client), the message
   address must be obtained using the appropriate operator. */

char *reference_message(enum component_role component)
{
  return (component == CLIENT_ROLE) ? "&" : "";
}

/* Return the naming prefix for structure access to a message. */

char *structure_prefix(enum specifier direction)
{
  return direction == IN_PARAMETER ? "in" : "out";
}

/* Return whether message access for the given component role and specifier (or
   direction) involves writing to (or populating) the message. */

int writing_to_message(enum component_role component, enum specifier direction)
{
  return (component == CLIENT_ROLE) && (direction & IN_PARAMETER) ||
         (component == SERVER_ROLE) && (direction & OUT_PARAMETER);
}
