/*
 * Main program.
 *
 * Copyright (C) 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "common.h"
#include "program.h"

/* Parser functions and settings. */

extern void yyparse(void);
extern void yyrestart(FILE *);
extern int yydebug;

/* Specialised parser functions. */

extern void reset(void);

/* Current filename and output options. */

char *output_prefix = NULL;
char *output_dir = NULL;

/* Compound interface and dispatcher output. */

char *output_compound = NULL;
char *output_compound_name = NULL;

int output_client = 0;
int output_server = 0;

int output_headers = 0;
int output_interfaces = 0;
int output_routines = 0;

int output_verbose = 0;

enum language output_language = C_LANGUAGE;



/* Return a copy of the given string without any extension. */

static char *make_prefix(const char *s)
{
  char *copy = strdup(s);
  char *ext;

  if (copy == NULL)
    return NULL;

  ext = strrchr(copy, '.');

  if (ext != NULL)
    *ext = '\0';

  return copy;
}

static int get_compound_prefix_conflict(int argc, char *argv[])
{
  char *arg, *prefix, *name;
  int index, result;

  for (index = optind; index < argc; index++)
  {
    arg = argv[index];
    prefix = make_prefix(arg);
    result = 0;

    if (prefix != NULL)
    {
      name = make_basename(prefix);

      if (name != NULL)
      {
        result = strcasecmp(name, output_compound);
        free(name);
      }

      free(prefix);
    }

    if (!result)
      return index;
  }

  return 0;
}


/* Option definitions. */
/* NOTE: --help and --version to be added. */

static struct option long_options[] = {
/* long opt       following           var pointer short opt */
  {"all",         no_argument,        NULL,       'a'       },
  {"client",      no_argument,        NULL,       'c'       },
  {"comp",        required_argument,  NULL,       'C'       },
  {"dir",         required_argument,  NULL,       'd'       },
  {"headers",     no_argument,        NULL,       'h'       },
  {"interfaces",  no_argument,        NULL,       'i'       },
  {"language",    required_argument,  NULL,       'l'       },
  {"comp-name",   required_argument,  NULL,       'N'       },
  {"routines",    no_argument,        NULL,       'r'       },
  {"server",      no_argument,        NULL,       's'       },
  {"verbose",     no_argument,        NULL,       'v'       },
  {0,             0,                  0,          0         },
  };


/* Main program, parsing indicated files or standard input. */

int main(int argc, char *argv[])
{
  FILE *fp;
  char *arg;
  int option, index, output_selected_content = 0, output_selected_role = 0;

#if YYDEBUG == 1
  yydebug = 0;
#endif

  /* Accept various options. */

  while (1)
  {
    option = getopt_long(argc, argv, "acC:d:hil:N:rsv", long_options, NULL);
    if (option == -1)
      break;

    switch (option)
    {
      /* Generate everything. */
      case 'a':
        output_client = 1; output_server = 1;
        output_headers = 1; output_interfaces = 1; output_routines = 1;
        output_selected_content = 1; 
        output_selected_role = 1; 
        break;

      /* Select client output. */
      case 'c':
        output_client = 1;
        output_selected_role = 1;
        break;

      /* Select compound interface filename prefix and output. */
      case 'C':
        output_compound = optarg;
        if (output_compound_name == NULL)
          output_compound_name = output_compound;
        output_selected_role = 1;
        break;

      /* Set output directory. */
      case 'd':
        output_dir = optarg;
        break;

      /* Generate headers. */
      case 'h':
        output_headers = 1;
        output_selected_content = 1;
        break;

      /* Generate interface headers. */
      case 'i':
        output_interfaces = 1;
        output_selected_content = 1;
        break;

      /* Select output language. */
      case 'l':
        if (!strcasecmp(optarg, "c"))
          output_language = C_LANGUAGE;
        else if (!strcasecmp(optarg, "c++"))
          output_language = CPP_LANGUAGE;
        else
        {
          fprintf(stderr, "Output language not recognised: %s\n", optarg);
          return 1;
        }
        break;

      /* Select compound interface name and output. */
      case 'N':
        output_compound_name = optarg;
        if (output_compound == NULL)
          output_compound = output_compound_name;
        output_selected_role = 1;
        break;

      /* Generate routines (definitions). */
      case 'r':
        output_routines = 1;
        output_selected_content = 1;
        break;

      /* Select server output. */
      case 's':
        output_server = 1;
        output_selected_role = 1;
        break;

      /* Verbose reporting. */
      case 'v':
        output_verbose = 1;
        break;

      default:
        return 1;
    }
  }

  /* Without any output selection, enable all output of the given kind. */

  if (!output_selected_content)
  {
    output_headers = 1; output_interfaces = 1; output_routines = 1;
  }

  if (!output_selected_role)
  {
    output_client = 1; output_server = 1;
  }

  /* Check any compound interface prefix against named files. */

  if (output_compound != NULL)
  {
    if ((index = get_compound_prefix_conflict(argc, argv)))
    {
      fprintf(stderr, "Compound interface prefix %s conflicts with filename %s.\n",
                      output_compound, argv[index]);
      return 1;
    }
  }

  /* Begin generating any compound interface code. */

  begin_compound_output();

  /* Process standard input without any named files. */

  if (optind >= argc)
  {
    yyparse();
    reset();
  }

  /* Process named files. */

  for (index = optind; index < argc; index++)
  {
    arg = argv[index];

    /* Handle actual files. */

    output_prefix = make_prefix(arg);

    if (output_prefix != NULL)
    {
      fp = fopen(arg, "r");

      if (fp != NULL)
      {
        yyrestart(fp);
        yyparse();
        fclose(fp);

        /* Reset any specialised parser state. */

        reset();
      }

      free(output_prefix);
    }
  }

  /* End generating any compound interface code. */

  end_compound_output();

  return 0;
}
