/*
 * Code generation from interface descriptions.
 *
 * Copyright (C) 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include "client.h"
#include "common.h"
#include "declaration.h"
#include "dispatch.h"
#include "message.h"
#include "program.h"
#include "server.h"
#include "structure.h"
#include "summary.h"
#include "templates.h"

/* Program state set in the main program. */

extern char *output_prefix, *output_dir,
            *output_compound, *output_compound_name;

extern int output_client, output_server,
           output_headers, output_interfaces, output_routines,
           output_verbose;

extern enum language output_language;

/* Current filename details. */

char *output_basename = NULL, *output_dirname = NULL;

/* Client, server and common files. */

FILE *client_fp = NULL, *client_header_fp = NULL,
     *server_fp = NULL, *server_header_fp = NULL,
     *server_interface_fp = NULL,
     *common_fp = NULL;

/* Compound interface and dispatcher files. */

FILE *compound_dispatch_fp = NULL, *compound_dispatch_header_fp = NULL,
     *compound_interface_fp = NULL, *compound_interfaces_fp = NULL;

/* Processed interfaces. */

int processed_interfaces;

/* Maximum number of input items expected by an interface. */

int max_input_items = 0;



/* Return the first interface name as the output prefix. */

static char *get_output_prefix(struct interface *i)
{
  if (output_prefix != NULL)
    return output_prefix;

  if (i == NULL)
    return NULL;

  for (; i->tail != NULL; i = i->tail);
  return i->name;
}

/* Obtain the basename and directory name of generated output. */

static int set_output_location(struct interface *iface)
{
  /* Obtain a prefix for the output files. */

  output_prefix = get_output_prefix(iface);

  if (output_prefix == NULL)
    return 0;

  /* Obtain the basename of the output prefix. */

  output_basename = make_basename(output_prefix);

  if (output_basename == NULL)
    return 0;

  /* Obtain the directory name from the output prefix. */

  if (output_dir == NULL)
    output_dirname = make_dirname(output_prefix);
  else
    output_dirname = output_dir;

  if (output_dirname == NULL)
  {
    free(output_basename);
    return 0;
  }

  return 1;
}

static void write_includes_separator(struct include *inc, FILE *fp)
{
  if (inc != NULL)
    fputs("\n", fp);
}

static void write_signature_end(FILE *fp)
{
  fputs(";\n", fp);
}


/* Generate compound code output, if requested. */

void begin_compound_output(void)
{
  char *compound_dirname = output_dir != NULL ? output_dir : ".";
  char *s;

  if (output_compound == NULL)
    return;

  processed_interfaces = 0;

  if (output_headers)
  {
    compound_dispatch_header_fp = get_output_file(server_header_filename,
                                                  compound_dirname, output_compound);
    if (compound_dispatch_header_fp == NULL)
      goto fail;
  }

  if (output_interfaces)
  {
    compound_interface_fp = get_output_file(server_interface_filename,
                                            compound_dirname, output_compound);
    if (compound_interface_fp == NULL)
      goto fail;

    compound_interfaces_fp = get_output_file(compound_interfaces_filename,
                                             compound_dirname, output_compound);
    if (compound_interfaces_fp == NULL)
      goto fail;
  }

  if (output_routines)
  {
    s = (output_language == CPP_LANGUAGE) ? server_filename_cpp :
                                            server_filename_c;
    compound_dispatch_fp = get_output_file(s, compound_dirname, output_compound);
    if (compound_dispatch_fp == NULL)
      goto fail;
  }

  /* Emit prologues. */

  if (compound_dispatch_fp != NULL)
  {
    fprintf(compound_dispatch_fp, compound_dispatch_prologue, output_compound);

    /* Write the handler function. */

    write_compound_handler_signature(compound_dispatch_fp);
    fputs("\n", compound_dispatch_fp);
    fprintf(compound_dispatch_fp, compound_handle_function, output_compound_name);

    /* Write the dispatch function. */

    write_compound_dispatch_signature(compound_dispatch_fp);
    fputs("\n", compound_dispatch_fp);
    fputs(compound_dispatch_function_prologue, compound_dispatch_fp);
  }

  if (compound_dispatch_header_fp != NULL)
  {
    fprintf(compound_dispatch_header_fp, compound_dispatch_header_prologue,
            output_compound);

    /* Write the handler signature. */

    write_compound_handler_signature(compound_dispatch_header_fp);
    fputs(";\n", compound_dispatch_header_fp);

    /* Write the dispatch signature. */

    write_compound_dispatch_signature(compound_dispatch_header_fp);
    fputs(";\n\n", compound_dispatch_header_fp);
  }

  if (compound_interface_fp != NULL)
    fprintf(compound_interface_fp, compound_interface_prologue,
            output_compound, output_compound_name);

  if (compound_interfaces_fp != NULL)
    fputs(compound_interfaces_prologue, compound_interfaces_fp);

  return;

fail:
  end_compound_output();
}

void write_compound_output(struct interface *iface)
{
  char *protocol;
  int input_items;

  if (compound_dispatch_fp != NULL)
  {
    protocol = get_protocol(iface->attributes);

    /* Populate a function dispatching to each interface-level function where
       the same protocol applies to an entire interface. */

    if (protocol != NULL)
      fprintf(compound_dispatch_fp, dispatch_function_interface_case,
              protocol, iface->name);

    /* Or dispatch to each operation defined at this level. */

    else
      write_dispatcher_cases(iface->signatures, compound_dispatch_fp, iface);
  }

  if (compound_dispatch_header_fp != NULL)
  {
    /* Compute the maximum number of items expected by each interface. */

    input_items = get_max_input_items(iface->signatures);

    if (input_items > max_input_items)
      max_input_items = input_items;
  }

  /* Add this interface to the compound interface. */

  if (compound_interface_fp != NULL)
    write_compound_interface(iface);

  /* Add this interface's header to the compound interface includes. */

  if (compound_interfaces_fp != NULL)
    fprintf(compound_interfaces_fp, "#include \"%s_interface.h\"\n",
            output_basename);

  processed_interfaces++;
}

/* Generate includes for the server and common declarations from each input
   file. */

void write_compound_dispatch_include(void)
{
  if (compound_dispatch_header_fp != NULL)
  {
    fprintf(compound_dispatch_header_fp, "#include \"%s_server.h\"\n",
            output_basename);
    fprintf(compound_dispatch_header_fp, "#include \"%s_common.h\"\n",
            output_basename);
  }
}

void write_compound_dispatch_signature(FILE *fp)
{
  char *s = (output_language == CPP_LANGUAGE) ?
            compound_dispatch_function_signature_cpp :
            compound_dispatch_function_signature_c;

  fprintf(fp, s, output_compound_name, output_compound_name);
}

void write_compound_handler_signature(FILE *fp)
{
  char *s = (output_language == CPP_LANGUAGE) ?
            compound_handle_function_signature_cpp :
            compound_handle_function_signature_c;

  fprintf(fp, s, output_compound_name, output_compound_name);
}

/* Augment a compound interface class declaration. */

void write_compound_interface(struct interface *iface)
{
  char *sep = processed_interfaces ? ", " : " ";
  fprintf(compound_interface_fp, "%spublic %s", sep, iface->name);
}

void end_compound_output(void)
{
  /* Close the files. */

  if (compound_dispatch_fp != NULL)
  {
    fputs(compound_dispatch_epilogue, compound_dispatch_fp);
    fclose(compound_dispatch_fp);
  }

  if (compound_dispatch_header_fp != NULL)
  {
    fprintf(compound_dispatch_header_fp, compound_dispatch_header_epilogue,
            output_compound_name, max_input_items);
    fclose(compound_dispatch_header_fp);
  }

  if (compound_interface_fp != NULL)
  {
    fputs(compound_interface_epilogue, compound_interface_fp);
    fclose(compound_interface_fp);
  }

  if (compound_interfaces_fp != NULL)
    fclose(compound_interfaces_fp);
}



/* Generate files containing interface details. */

void write_files(struct interface *iface)
{
  char *server_filename;

  /* NOTE: Should exit with an error. */

  if (!set_output_location(iface))
    return;

  /* Include server details for a compound interface dispatch function. */

  write_compound_dispatch_include();

  /* Open the separate files and write the details of each interface. */

  if (output_headers)
  {
    if (output_client)
    {
      client_header_fp = get_output_file(client_header_filename, output_dirname, output_basename);
      if (client_header_fp == NULL)
        goto finalisation;
    }

    if (output_server)
    {
      server_header_fp = get_output_file(server_header_filename, output_dirname, output_basename);
      if (server_header_fp == NULL)
        goto finalisation;
    }

    common_fp = get_output_file(common_filename, output_dirname, output_basename);
    if (common_fp == NULL)
      goto finalisation;
  }

  if (output_interfaces)
  {
    if (output_server)
    {
      server_interface_fp = get_output_file(server_interface_filename, output_dirname, output_basename);
      if (server_interface_fp == NULL)
        goto finalisation;
    }
  }

  if (output_routines)
  {
    if (output_client)
    {
      client_fp = get_output_file(client_filename, output_dirname, output_basename);
      if (client_fp == NULL)
        goto finalisation;
    }

    if (output_server)
    {
      server_filename = (output_language == CPP_LANGUAGE) ? server_filename_cpp : server_filename_c;
      server_fp = get_output_file(server_filename, output_dirname, output_basename);
      if (server_fp == NULL)
        goto finalisation;
    }
  }

  /* Emit prologues. */

  if (client_fp != NULL)
    fprintf(client_fp, client_prologue, output_basename, output_basename);

  if (client_header_fp != NULL)
    fputs(header_prologue, client_header_fp);

  if (server_fp != NULL)
    fprintf(server_fp, server_prologue, output_basename, output_basename);

  if (server_header_fp != NULL)
    fputs(server_header_prologue, server_header_fp);

  if (server_interface_fp != NULL)
    fputs(header_prologue, server_interface_fp);

  if (common_fp != NULL)
    fputs(header_prologue, common_fp);

  /* Write the interface details. */

  write_interfaces(iface);

finalisation:

  /* Close the files. */

  if (client_fp != NULL)
    fclose(client_fp);

  if (client_header_fp != NULL)
    fclose(client_header_fp);

  if (server_fp != NULL)
    fclose(server_fp);

  if (server_header_fp != NULL)
    fclose(server_header_fp);

  if (server_interface_fp != NULL)
    fclose(server_interface_fp);

  if (common_fp != NULL)
    fclose(common_fp);

  /* Free the names. */

  if (output_dir == NULL)
    free(output_dirname);

  free(output_basename);
}

/* Generate the source file corresponding to the interface. */

void write_interfaces(struct interface *iface)
{
  if (iface == NULL)
    return;

  /* The list is reversed. */

  write_interfaces(iface->tail);

  /* Write this particular interface after all referenced ones. */

  if (output_verbose)
    show_interface(iface);

  /* Record details of this interface in the compound output. */

  write_compound_output(iface);

  /* Emit function definitions. */

  if (client_fp != NULL)
    write_functions(iface->signatures, client_fp, iface,
                    write_client_function);

  if (server_fp != NULL)
  {
    write_functions(iface->signatures, server_fp, iface,
                    write_server_function);

    write_dispatcher(iface->signatures, server_fp, iface);
  }

  /* Emit the headers. */

  if (client_header_fp != NULL)
  {
    /* Emit include statements. */

    write_includes_separator(iface->includes, client_header_fp);
    write_includes(iface->includes, client_header_fp);

    /* Emit signatures. */

    write_signatures(iface->signatures, client_header_fp, iface,
                     write_client_signature);
  }

  if (server_header_fp != NULL)
  {
    /* Emit interface include where the type is employed by signatures. */

    if (output_language == CPP_LANGUAGE)
      fprintf(server_header_fp, "#include \"%s_interface.h\"\n", output_basename);

    /* Emit signatures. */

    write_signatures(iface->signatures, server_header_fp, iface,
                     write_server_signature);

    write_dispatcher_signature(server_header_fp, iface);
  }

  if (server_interface_fp != NULL)
  {
    /* Emit include statements. */

    write_includes_separator(iface->includes, server_interface_fp);
    write_includes(iface->includes, server_interface_fp);

    /* Begin any namespace. */

    if (output_language == CPP_LANGUAGE)
      fprintf(server_interface_fp, server_interface_prologue_cpp, iface->name);

    /* Emit signatures. */

    write_signatures(iface->signatures, server_interface_fp, iface,
                     write_server_interface_signature);

    /* Terminate any namespace. */

    if (output_language == CPP_LANGUAGE)
      fputs(server_interface_epilogue_cpp, server_interface_fp);
  }

  /* Emit common structures. */

  if (common_fp != NULL)
  {
    /* Emit include statements. */

    write_includes_separator(iface->includes, common_fp);
    write_includes(iface->includes, common_fp);

    /* Emit opcodes. */

    write_opcode_definition(iface, common_fp);

    /* Emit message access structures. */

    write_structures(iface->signatures, common_fp, iface);
  }
}

/* Write include statements. */

void write_includes(struct include *inc, FILE *fp)
{
  if (inc == NULL)
    return;

  /* The list is reversed. */

  write_includes(inc->tail, fp);

  fprintf(fp, "#include %s\n", inc->filename);
}

/* Generate signatures. */

void write_signatures(struct signature *sig, FILE *fp, struct interface *iface,
                      void (*write_signature)(struct signature *, FILE *, struct interface *))
{
  if (sig == NULL)
    return;

  write_signature(sig, fp, iface);
  write_signatures(sig->tail, fp, iface, write_signature);
}

/* Generate functions corresponding to the signatures. */

void write_functions(struct signature *sig, FILE *fp, struct interface *iface,
                     void (*write_function)(struct signature *, FILE *, struct interface *))
{
  if (sig == NULL)
    return;

  write_function(sig, fp, iface);
  write_functions(sig->tail, fp, iface, write_function);
}
