/*
 * Server code generation.
 *
 * Copyright (C) 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include "common.h"
#include "declaration.h"
#include "message.h"
#include "server.h"
#include "templates.h"



/* Output language, set in the main program. */

extern enum language output_language;



static void write_signature_end(FILE *fp)
{
  fputs(";\n", fp);
}

/* Generate operation wrapper function details used by the server. */

void _write_server_signature(struct signature *sig, FILE *fp, struct interface *iface)
{
  /* NOTE: Qualifier discarded for now. */

  char *opname = get_operation_name(iface, sig),
       *type = (output_language == CPP_LANGUAGE) ? iface->name : "void";

  fprintf(fp, "\nlong wrap_%s(ipc_message_t *msg, %s *_self)", opname, type);

  free(opname);
}

/* Generate each signature for a wrapper function used by the server. */

void write_server_signature(struct signature *sig, FILE *fp, struct interface *iface)
{
  _write_server_signature(sig, fp, iface);
  write_signature_end(fp);
}

/* Generate each operation signature or "function prototype" for server use. */

void write_server_interface_signature(struct signature *sig, FILE *fp, struct interface *iface)
{
  char *opname = get_interface_operation_name(iface, sig, output_language);

  /* NOTE: Qualifier discarded for now. */

  int have_init_param = (output_language == C_LANGUAGE);

  /* Indent and provide a qualifier for C++ class declarations. */

  char *type = (output_language == CPP_LANGUAGE) ? iface->name : "void",
       *qualifier = (output_language == CPP_LANGUAGE) ? "  virtual " : "",
       *value_qualifier = (output_language == CPP_LANGUAGE) ? " = 0" : "",
       *init_param = have_init_param ? "void *_self" : "";

  /* Introduce an initial parameter for the component state. */

  fprintf(fp, "\n%slong %s(%s", qualifier, opname, init_param);

  if (have_init_param)
    write_parameter_separator(sig->parameters, fp);
  write_parameters(sig->parameters, fp, SIGNATURE_ROLE);

  fprintf(fp, ")%s;\n", value_qualifier);

  free(opname);
}

/* Generate function source code for each operation, with the generated function
   unpacking a message, calling the actual operation and repacking the
   results. */

void write_server_function(struct signature *sig, FILE *fp, struct interface *iface)
{
  char *type = (output_language == CPP_LANGUAGE) ? iface->name : "void";

  _write_server_signature(sig, fp, iface);
  fputs("\n{\n", fp);
  write_server_function_body(sig->parameters, fp, iface, sig);
  fputs("}\n", fp);
}

/* Generate a function body corresponding to an operation for server use. */

void write_server_function_body(struct parameter *param, FILE *fp,
                                struct interface *iface, struct signature *sig)
{
  int input_words, input_items, output_words, output_items;
  char *opname = get_operation_name(iface, sig);

  /* Generate the prologue. */

  fputs(server_function_body_prologue, fp);

  /* Count the number of input words, items and output words. */

  count_parameters(param, &input_words, &input_items, &output_words, &output_items);

  /* Generate variable declarations. */

  write_declarations(param, ANY_PARAMETER, ANY_CLASS, fp);

  /* Generate types parameterised with the qualified operation name. */

  write_accessor_declaration(input_words, output_words, opname, fp);

  /* Unpack each word and item from the message into variables. */

  if (input_words)
  {
    write_accessor_initialisation(IN_PARAMETER, SERVER_ROLE, opname, fp);
    write_message_access(param, SERVER_ROLE, IN_PARAMETER, WORD_CLASS, fp);
  }

  if (input_items)
  {
    fputs("\n", fp);
    write_message_access(param, SERVER_ROLE, IN_PARAMETER, ITEM_CLASS, fp);
  }

  /* Invoke the actual operation using the variables. */

  write_server_function_call(param, fp, iface, sig);

  /* Pack the outputs into the message and perform any other housekeeping. */

  if (output_words)
  {
    write_accessor_initialisation(OUT_PARAMETER, SERVER_ROLE, opname, fp);
    write_message_access(param, SERVER_ROLE, OUT_PARAMETER, WORD_CLASS, fp);
  }

  if (output_items)
  {
    fputs("\n", fp);
    write_message_access(param, SERVER_ROLE, OUT_PARAMETER, ITEM_CLASS, fp);
  }

  /* Return success. */

  fprintf(fp, "\n  return L4_EOK;\n");

  /* Free allocated strings. */

  free(opname);
}

/* Generate an invocation of the actual server operation. */

void write_server_function_call(struct parameter *param, FILE *fp,
                                struct interface *iface, struct signature *sig)
{
  char *name, *addr,
       *opname = get_interface_operation_name(iface, sig, output_language);

  fputs("\n  err = ", fp);

  /* Access a method when emitting C++. */

  if (output_language == CPP_LANGUAGE)
    fprintf(fp, "_self->%s(", opname);

  /* Employ a function otherwise. */

  else
  {
    fprintf(fp, "%s(_self", opname);
    write_parameter_separator(param, fp);
  }

  /* Generate the parameter list, employing addresses for output parameters. */

  for (; param != NULL; param = param->tail)
  {
    addr = param->specifier & OUT_PARAMETER ? "&" : "";
    name = get_parameter_name(param->identifiers);
    fprintf(fp, "%s%s", addr, name);
    write_parameter_separator(param->tail, fp);
  }

  fputs(");\n", fp);

  /* Emit post-invocation details. */

  fputs(server_function_body_call, fp);

  /* Free allocated strings. */

  free(opname);
}
